﻿using System.ComponentModel;
using GraphicsManager.Enums;
using GraphicsManager.Interfaces;
using GraphicsManager.Objects;
using GraphicsManager.Objects.Core;
using OpenTK.Graphics.OpenGL4;
using OpenTK.Mathematics;
using OpenTK.Windowing.Common;
using OpenTK.Windowing.Common.Input;
using OpenTK.Windowing.Desktop;
using OpenTK.Windowing.GraphicsLibraryFramework;

namespace GraphicsManager;

public class FPSWindow : GameWindow , IWindow
{
	public TextureManager TextureManager { get; private set; }
	private TitleBar? _tb = null;
	
	public bool ShowMissingChar { get; set; }

	public TitleBar? CustomeTitleBar
	{
		get => _tb;
		set
		{
			_tb = value;
			//some setting/changeing code
		}
	}

	public IRenderObject? focused { get; set; }
	public void CenterWindow(int mon)
	{
		Box2i clientArea = Monitors.GetMonitors()[mon].ClientArea;
		int num = (clientArea.Min.X + clientArea.Max.X - Size.X) / 2;
		int num2 = (clientArea.Min.Y + clientArea.Max.Y - Size.Y) / 2;
		if (num < clientArea.Min.X)
		{
			num = clientArea.Min.X;
		}

		if (num2 < clientArea.Min.Y)
		{
			num2 = clientArea.Min.Y;
		}

		ClientRectangle = new Box2i(num, num2, num + Size.X, num2 + Size.Y);
	}

	public IParent? Parent { get; } = null;
	public MouseCursor HoverMouse { get; set; } = MouseCursor.Default;
	public Vector2 LocationAsFloat { get; } = new Vector2(0f, 0f);
	public FPSWindow(NativeWindowSettings nativeWindowSettings, GameWindowSettings gws) : base(gws,nativeWindowSettings)
	{
		TextureManager = TextureManager.GetTextureManager(Context, this);
		Context.MakeCurrent();
		if (!Texture.TextureShader.ContainsKey(Context)) Texture.TextureShader.Add(Context, new("RectangleTexture", true, Texture:true));
		if (!Rectangle.DefaultShader.ContainsKey(Context)) Rectangle.DefaultShader.Add(Context, new("Rectangle", true));
		if (!Rectangle.DefaultAlphaShader.ContainsKey(Context)) Rectangle.DefaultAlphaShader.Add(Context, new("AlphaChannel", true, Texture:true));
		if (!Rectangle.DefaultAlphaTextureShader.ContainsKey(Context)) Rectangle.DefaultAlphaTextureShader.Add(Context, new("AlphaChannelTexture", true, Texture:true));
		if (!Label.DefaultTextShader.ContainsKey(Context)) Label.DefaultTextShader.Add(Context, new("Label", true));
		if (!Label._characters.ContainsKey(Context)) Label._characters.Add(Context, new());
		last = WindowState;
		KeyDown += OnKeyDownn;
	}
	
	private WindowState last;
	private bool fs = false;
	private Vector2i os, loc;
	public bool CustomF11 = true;
	private unsafe void OnKeyDownn(KeyboardKeyEventArgs obj)
	{
		if (obj.Key != Keys.F11 || WindowBorder == WindowBorder.Fixed) return;
		if (CustomF11)
		{
			GLFW.WindowHint(WindowHintBool.AutoIconify, true);
			if (WindowState != WindowState.Normal) last = WindowState;
			switch (fs)
			{
				case false:
					loc = Location;
					os = Size;
					WindowBorder = WindowBorder.Hidden;
					Location = new(4090, 0);
					fs = true;
					break;
				case true:
					WindowBorder = WindowBorder.Resizable;
					GLFW.RestoreWindow(this.WindowPtr);
					Size = os;
					Location = loc;
					fs = false;
					break;
			}
		}
		else
		{
			switch (fs)
			{
				case false:
					WindowState = WindowState.Fullscreen;
					fs = true;
					break;
				case true:
					WindowState = WindowState.Normal;
					fs = false;
					break;
			}
		}
	}

	public FPSWindow() : this(new NativeWindowSettings(), new GameWindowSettings())
	{
	}
	public Vector3i Position { get; } = new(0);
	public Color4 BackgroundColor { get; set; } = new Color4(0, 0, 0, 255);

	public ControlList Controls { get; } = new();
#region Cool Math Things
	public float[] RctToFloat(int x, int y, int Width, int Height, bool hasTexture = false, float z = 0.0f)
	{
		if (hasTexture)
		{
			return new float[20] {
				IntToFloat(x + Width), IntToFloat(y, true), z, 1.0f, 1.0f,// top r
                IntToFloat(x + Width), IntToFloat(y + Height, true), z, 1.0f, 0.0f,//b r
                IntToFloat(x), IntToFloat(y + Height, true), z, 0.0f, 0.0f,//bot l
                IntToFloat(x), IntToFloat(y, true), z, 0.0f, 1.0f// top l
            };
		}
		else
		{
			return new float[12] {
				IntToFloat(x + Width), IntToFloat(y, true), z,// top r
                IntToFloat(x + Width), IntToFloat(y + Height, true), z, //b r
                IntToFloat(x), IntToFloat(y + Height, true), z, //bot l
                IntToFloat(x), IntToFloat(y, true), z,// top l
            };
		}
	}

	public Vector3 PointToVector(float x, float y, float z = 0.0f)
	{
		return new Vector3(IntToFloat(x), IntToFloat(y, true), z);
	}

	public float IntToFloat(float p, bool Invert = false)
	{
		float Size = (Invert ? this.Size.Y : this.Size.X);
		double half = Math.Round((double)Size / (double)2, 1);
		double Per = Math.Round((double)1 / half, 15);
		if (p == half) return 0.0f;
		if (Invert)
		{
			if (p > half) return (float)(((double)(p - half) * Per) * -1);
			else return (float)(1 - (p * Per));
		}
		else
		{
			if (p > half) return (float)((double)(p - half) * Per);
			else return (float)((1 - (p * Per)) * -1);
		}
	}

	public float FloatToInt(float p, bool Invert = false)
	{
		float Size = (Invert ? this.Size.Y : this.Size.X);
		double half = Math.Round((double)Size / (double)2, 15);
		if (p == 0) return (int)half;
		if (Invert)
		{
			if (p < 0)
			{
				p *= -1;
				p++;
				return (float)(half * p);
			}
			else
			{
				return (float)(half - (p * half));
			}
		}
		else
		{
			if (p < 0)
			{
				p *= -1;
				p++;
				return (float)(Size - (half * p));
			}
			else
			{
				return (float)(p * half + half);
			}
		}
	}
#endregion

	public void ForceUpdate(ResizeEventArgs e)
	{
		BlockDraw = true;
		for (int i = 0; i < Controls.Length; i++)
		{
			if (!Controls[i].Loaded) continue;
			bool top = (Controls[i].Anchor & ObjectAnchor.Top) == ObjectAnchor.Top;
			bool left = (Controls[i].Anchor & ObjectAnchor.Left) == ObjectAnchor.Left;
			bool right = (Controls[i].Anchor & ObjectAnchor.Right) == ObjectAnchor.Right;
			bool bottom = (Controls[i].Anchor & ObjectAnchor.Bottom) == ObjectAnchor.Bottom;
			if (!top && !bottom) { Controls[i].Anchor |= ObjectAnchor.Top; top = true; }
			if (!left && !right) { Controls[i].Anchor |= ObjectAnchor.Left; left = true; }
			int lx = (left ? Controls[i].Location.X : Size.X - Controls[i].Distance.X - Controls[i].Size.X);
			int ly = (top ? Controls[i].Location.Y : Size.Y - Controls[i].Distance.Y - Controls[i].Size.Y);
			int sy = (bottom ? Size.Y - Controls[i].Distance.Y - ly : Controls[i].Size.Y);
			int sx = (right ? Size.X - Controls[i].Distance.X - lx : Controls[i].Size.X);
			Controls[i].Size = new(sx, sy);
			Controls[i].Location = new(lx, ly, Controls[i].Location.Z);
			if (Controls[i] is IParent parent)
			{
				parent.ParentResize(e);
			}
		}
		BlockDraw = false;
	}

	public void ParentResize(ResizeEventArgs e)
	{
		base.OnResize(e);
		if (e.Width == 0 && e.Height == 0 && WindowState != WindowState.Fullscreen) return;
		GL.Viewport(0, 0, e.Width, e.Height); 
		Matrix4.CreateOrthographicOffCenter(0.0f, 800.0f, 0.0f, 600.0f, 0.1f, 100.0f);
		ForceUpdate(e);
	}

	private int frame = 0;

	public void ReportSizeUpdate(IRenderObject Control)
	{
		
	}

	protected override void OnResize(ResizeEventArgs e)
	{
		ParentResize(e);
	}

	protected override void OnClosing(CancelEventArgs e)
	{
		Context.MakeCurrent();
		GL.BindBuffer(BufferTarget.ArrayBuffer, 0);
		GL.BindVertexArray(0);
		GL.UseProgram(0);
		for (int i = 0; i < Controls.Length; i++)
		{
			Controls[i].Clean();
		}
		base.OnClosing(e);
	}

	

	public Rendertype Rendertype { get; set; } = Rendertype.ControlUpdates;

	public bool CanControleUpdate { get; set; } = true;

	public int FPS { get; set; } = 0;
	public event Func<FPSWindow, Task>? WindowLoaded;
	public List<Window> Windows = new();

	public IRenderObject? HoveringControl { get; set; }

	private void WhenMouseMove(MouseMoveEventArgs obj)
	{
		IRenderObject? cb = null;
		IRenderObject? checkcontrol(IRenderObject render, int add_x = 0, int addy = 0, IRenderObject? CurrentBest = null)
		{
			void ClearHoverChain(IRenderObject ChainObject)
			{
				if (ChainObject is IParent parent)
				{
					for (int i = 0; i < parent.Controls.Length; i++)
					{
						ClearHoverChain(parent.Controls._internal[i]);
					}
				}

				if (ChainObject.AllowHoverFromBehind) ChainObject.MouseInside = false;
			}

			if (!render.Visible ||
			    render.IgnoreHover ||
			    render.Location.X + add_x > obj.X ||
			    render.Location.Y + addy > obj.Y ||
			    render.Location.X + render.Size.X + add_x < obj.X ||
			    render.Location.Y + render.Size.Y + addy < obj.Y)
			{
				ClearHoverChain(render);
				return CurrentBest;
			}
			if (render.AllowHoverFromBehind) render.MouseInside = true;
			if (render is IParent p)
			{
				for (int i = p.Controls.Length - 1; i >= 0; i--)
				{
					CurrentBest = checkcontrol(p.Controls._internal[i], add_x + render.Location.X, addy + render.Location.Y, CurrentBest);
				}
			}
			if (CurrentBest is null)
			{
				CurrentBest = render;
			}
			else
			{
				if (CurrentBest.Location.Z < render.Location.Z) CurrentBest = render;
			}

			return CurrentBest;
		}
		for (int i = Controls.Length - 1; i >= 0; i--)
		{
			cb = checkcontrol(Controls._internal[i],0,0,cb);
		}

		if (cb != HoveringControl)
		{
			IRenderObject? old = HoveringControl;
			HoveringControl = cb;
			if (cb is not null)
			{
				cb.MouseInside = true;
				Cursor = cb.HoverMouse;
			}
			else Cursor = HoverMouse;
			if (old is not null) old.MouseInside = false;
		}
	}
	
	public bool BlockDraw { get; set; }

	public void TryDraw()
	{
		
	}
	
	private int initthread;

	public bool InvokeRequired
	{
		get
		{
			return initthread != Thread.CurrentThread.ManagedThreadId;
		}
	}

	private Queue<Action> invokes = new();
	
	public void Invoke(Action A)
	{
		invokes.Enqueue(A);
	}
	
	protected override void OnRenderFrame(FrameEventArgs args)
	{
		Context.MakeCurrent();
		frame++;
		Console.WriteLine($"Drawing Frame: {frame}");
		GL.Enable(EnableCap.Blend);
		GL.BlendFunc(BlendingFactor.SrcAlpha, BlendingFactor.OneMinusSrcAlpha);
		GL.BlendFunc(0, BlendingFactorSrc.SrcAlpha, BlendingFactorDest.OneMinusSrcAlpha);
		GL.Enable(EnableCap.ScissorTest);
		
		GL.ClearColor(BackgroundColor.R, BackgroundColor.G, BackgroundColor.B, (BackgroundColor.A * -1) + 1);
		if (invokes.Any())
		{
			for (int i = 0; i < invokes.Count; i++) invokes.Dequeue().Invoke();
		}
		IEnumerable<IRenderObject> needload = Controls.Where(a => a.Loaded == false);
		
		if (needload.Any())
		{
			BlockDraw = true;
			foreach (IRenderObject obj in needload)
			{
				obj.LoadToParent(this, this);
			}
			BlockDraw = false;
		}
		
		GL.Scissor(0, 0, Size.X, Size.Y);
		GL.Clear(ClearBufferMask.ColorBufferBit | ClearBufferMask.DepthBufferBit);
		for (int i = 0; i < Controls.Length; i++)
		{
			if (!Controls[i].Loaded) continue;
			GL.Scissor(0, 0, Size.X, Size.Y);
			Controls[i].Draw(0,0,Size.X, Size.Y);
		}
		Context.SwapBuffers();
		base.OnRenderFrame(args);
	}
}