﻿using GraphicsManager.Structs;
using OpenTK.Mathematics;
using OpenTK.Windowing.Desktop;
using SharpFont;
using SixLabors.ImageSharp;
using SixLabors.ImageSharp.Formats.Bmp;
using SixLabors.ImageSharp.Formats.Gif;
using SixLabors.ImageSharp.Formats.Jpeg;
using SixLabors.ImageSharp.Formats.Pbm;
using SixLabors.ImageSharp.Formats.Png;
using SixLabors.ImageSharp.Formats.Tga;
using SixLabors.ImageSharp.Formats.Tiff;
using SixLabors.ImageSharp.Formats.Webp;
using SixLabors.ImageSharp.PixelFormats;
using SixLabors.ImageSharp.Processing;
using GL = OpenTK.Graphics.OpenGL4.GL;
using Image = SixLabors.ImageSharp.Image;
using PixelFormat = OpenTK.Graphics.OpenGL4.PixelFormat;
using PixelInternalFormat = OpenTK.Graphics.OpenGL4.PixelInternalFormat;
using PixelStoreParameter = OpenTK.Graphics.OpenGL4.PixelStoreParameter;
using PixelType = OpenTK.Graphics.OpenGL4.PixelType;
using TextureMagFilter = OpenTK.Graphics.OpenGL4.TextureMagFilter;
using TextureMinFilter = OpenTK.Graphics.OpenGL4.TextureMinFilter;
using TextureParameterName = OpenTK.Graphics.OpenGL4.TextureParameterName;
using TextureTarget = OpenTK.Graphics.OpenGL4.TextureTarget;
using TextureUnit = OpenTK.Graphics.OpenGL4.TextureUnit;
using TextureWrapMode = OpenTK.Graphics.OpenGL4.TextureWrapMode;

namespace GraphicsManager.Objects.Core;

public class Texture
{
	public static readonly Dictionary<IGLFWGraphicsContext, Shader> TextureShader = new ();

	private static Configuration imgcfg = new(
		new PngConfigurationModule(),
		new JpegConfigurationModule(),
		new GifConfigurationModule(),
		new BmpConfigurationModule(),
		new PbmConfigurationModule(),
		new TgaConfigurationModule(),
		new TiffConfigurationModule(),
		new WebpConfigurationModule())
	{
		PreferContiguousImageBuffers = true
	};

	public int handel;
	public Vector2i? RawSize = null;
	public int Location { get; set; } = 1;
	public Vector2 MaxText = new(1f);

	private Texture(byte[] pixels, int Width, int Height)
	{
		handel = GL.GenTexture();
		RawSize = new(Width, Height);
		GL.BindTexture(TextureTarget.Texture2D, handel);
		GL.TexImage2D(TextureTarget.Texture2D, 0, PixelInternalFormat.Rgba, Width, Height, 0, PixelFormat.Rgba, PixelType.UnsignedByte, pixels);
		GL.TexParameter(TextureTarget.Texture2D, TextureParameterName.TextureMinFilter, (int)TextureMinFilter.Linear);
		GL.TexParameter(TextureTarget.Texture2D, TextureParameterName.TextureMagFilter, (int)TextureMagFilter.Linear);
		GL.TexParameter(TextureTarget.Texture2D, TextureParameterName.TextureWrapS, (int)TextureWrapMode.Repeat);
		GL.TexParameter(TextureTarget.Texture2D, TextureParameterName.TextureWrapT, (int)TextureWrapMode.Repeat);
	}
	
	private Texture(Image<Rgba32> image, Memory<Rgba32> m)
	{
		handel = GL.GenTexture();
		GL.BindTexture(TextureTarget.Texture2D, handel);
		RawSize = new(image.Width, image.Height);
		GL.TexImage2D(TextureTarget.Texture2D, 0, PixelInternalFormat.Rgba, image.Width, image.Height, 0, PixelFormat.Rgba, PixelType.UnsignedByte, ref m.Span[0]);
		GL.TexParameter(TextureTarget.Texture2D, TextureParameterName.TextureMinFilter, (int)TextureMinFilter.Linear);
		GL.TexParameter(TextureTarget.Texture2D, TextureParameterName.TextureMagFilter, (int)TextureMagFilter.Linear);
		GL.TexParameter(TextureTarget.Texture2D, TextureParameterName.TextureWrapS, (int)TextureWrapMode.Repeat);
		GL.TexParameter(TextureTarget.Texture2D, TextureParameterName.TextureWrapT, (int)TextureWrapMode.Repeat);
	}
	
	
	internal static Texture GetTexture(Stream File)
	{
		try
		{
			Image<Rgba32> image = Image.Load<Rgba32>(imgcfg, File);
			image.Mutate(x =>
			{
				x.Flip(FlipMode.Vertical);
			});
			byte[] pixels = new byte[4 * image.Width * image.Height];
			image.CopyPixelDataTo(pixels);
			Texture t = new(pixels, image.Width, image.Height);
			pixels = Array.Empty<byte>();
			image.Dispose();
			return t;
		}
		catch (Exception e)
		{
			Console.WriteLine(e);
			throw;
		}
	}
	
	internal static Texture GetTexture(Stream File, TextureManager tm, bool Missing)
	{
		try
		{
			Image<Rgba32> image = Image.Load<Rgba32>(imgcfg, File);
			image.Mutate(x =>
			{
				x.Flip(FlipMode.Vertical);
			});
			File.Position = 0;
			Texture? t;
			if (image.DangerousTryGetSinglePixelMemory(out Memory<Rgba32> m))
			{
				t = new(image, m);
				image.Dispose();
				m = null;
			}
			else
			{
				_ = image.DangerousTryGetSinglePixelMemory(out Memory<Rgba32> mm);
				if (Missing)
				{
					File.Position = 0;
					Console.WriteLine("Getting texture from stream failed, switching to Missing.png.");
					image.Dispose();
					t = GetTexture(Tools.GetResourceStream("GraphicsManager.Resources.Textures.Missing.png"), tm, false);
				}
				else
				{
					File.Position = 0;
					Console.WriteLine("Getting texture from stream failed, switching to fallback.");
					t = GetTexture(File);
				}
				
			}
			return t;
		}
		catch (Exception e)
		{
			Console.WriteLine(e);
			throw;
		}
	}

	private Texture()
	{
	}

	internal static Character GetChar(Face face, char charter)
	{
		ushort temp = (charter);
		if (face.GetCharIndex(temp) == 0)
		{
			return new()
			{
				Size = new Vector2(0, 0),
				Bearing = new Vector2(0, 0),
				Advance = 0,
				AdvanceY = ((Math.Abs(face.Descender) / (double)face.UnitsPerEM) * 16)
			};
		}
		face.LoadChar(temp, LoadFlags.Render, LoadTarget.Normal);
		GlyphSlot glyph = face.Glyph;
		FTBitmap bitmap = glyph.Bitmap;
				
		return new()
		{
			Size = new Vector2(bitmap.Width, bitmap.Rows),
			Bearing = new Vector2(glyph.BitmapLeft, glyph.BitmapTop),
			Advance = glyph.Advance.X.Value,
			AdvanceY = ((Math.Abs(face.Descender) / (double)face.UnitsPerEM) * 16)
		};
	}
	
	internal static Character GetChar(FontInteraction l, char charter, IGLFWGraphicsContext con)
	{
		try
		{
			if (!Label._characters[con].ContainsKey(l)) Label._characters[con].Add(l, new Dictionary<uint, Character>());
			if (Label._characters[con][l].ContainsKey(charter)) return Label._characters[con][l][(ushort)charter];
			for (int i = 0; i < l.CurrentFonts.Length; i++)
			{
				try
				{
					l.CurrentFonts[i].Face.SetPixelSizes(0, l.PixelHeight);
					l.CurrentFonts[i].Face.SelectCharmap(Encoding.Unicode);
				}
				catch
				{
					continue;
				}
				ushort temp = (charter);
				if (l.CurrentFonts[i].Face.GetCharIndex(temp) == 0) continue;

				return GetChar(l.CurrentFonts[i].Face, charter);
			}
		}
		catch (Exception e)
		{
			Console.WriteLine(e);
			throw;
		}
		return GetChar(l.CurrentFonts[0].Face, (char)1);
	}

	internal static Character GetChar(FontInteraction l, char charter)
	{
		try
		{
			for (int i = 0; i < l.CurrentFonts.Length; i++)
			{
				try
				{
					l.CurrentFonts[i].Face.SetPixelSizes(0, l.PixelHeight);
					l.CurrentFonts[i].Face.SelectCharmap(Encoding.Unicode);
				}
				catch
				{
					continue;
				}
				ushort temp = (charter);
				if (l.CurrentFonts[i].Face.GetCharIndex(temp) == 0) continue;

				return GetChar(l.CurrentFonts[i].Face, charter);
			}
		}
		catch (Exception e)
		{
			Console.WriteLine(e);
			throw;
		}
		return GetChar(l.CurrentFonts[0].Face, (char)1);
	}

	private static Dictionary<FontInteraction, Texture> BadChars = new();


	internal static Texture TextureForChar(IGLFWGraphicsContext con, FontInteraction l, char charter, Shader s)
	{
		try
		{
			ushort temp = ((ushort)charter);
			for (int i = 0; i < l.CurrentFonts.Length; i++)
			{
				if (!Label._characters[con].ContainsKey(l)) Label._characters[con].Add(l, new Dictionary<uint, Character>());
				if (Label._characters[con][l].ContainsKey(charter)) return Label._characters[con][l][(ushort)charter].Texture;
				try
				{
					l.CurrentFonts[i].Face.SetPixelSizes(0, l.PixelHeight);
				}
				catch (Exception e)
				{
					Console.WriteLine("should never happen");
					continue;
				}
				l.CurrentFonts[i].Face.SelectCharmap(Encoding.Unicode);
				
				if (l.CurrentFonts[i].Face.GetCharIndex(temp) == 0) continue;
				l.CurrentFonts[i].Face.LoadChar(temp, LoadFlags.Render, LoadTarget.Normal);
				GlyphSlot glyph = l.CurrentFonts[i].Face.Glyph;
				FTBitmap bitmap = glyph.Bitmap;
				Texture t = new();
				t.Unit = (s.GetUniformLocation("u_texture") switch
				{
					0 => TextureUnit.Texture0,
					1 => TextureUnit.Texture1,
					2 => TextureUnit.Texture2,
					3 => TextureUnit.Texture3,
					4 => TextureUnit.Texture4,
					5 => TextureUnit.Texture5,
					6 => TextureUnit.Texture6,
					7 => TextureUnit.Texture7,
					8 => TextureUnit.Texture8,
					9 => TextureUnit.Texture9,
				});
				GL.PixelStore(PixelStoreParameter.UnpackAlignment, 1);
				GL.ActiveTexture(t.Unit);
				t.handel = GL.GenTexture();
				GL.BindTexture(TextureTarget.Texture2D, t.handel);
				GL.TexImage2D(TextureTarget.Texture2D, 0,
					PixelInternalFormat.R8, bitmap.Width, bitmap.Rows, 0,
					PixelFormat.Red, PixelType.UnsignedByte, bitmap.Buffer);
				GL.TextureParameter(t.handel, TextureParameterName.TextureMinFilter, (int)TextureMinFilter.Linear);
				GL.TextureParameter(t.handel, TextureParameterName.TextureMagFilter, (int)TextureMagFilter.Linear);
				GL.TextureParameter(t.handel, TextureParameterName.TextureWrapS, (int)TextureWrapMode.ClampToEdge);
				GL.TextureParameter(t.handel, TextureParameterName.TextureWrapT, (int)TextureWrapMode.ClampToEdge);

				Character cha = new()
				{
					Size = new Vector2(bitmap.Width, bitmap.Rows),
					Bearing = new Vector2(glyph.BitmapLeft, glyph.BitmapTop),
					Advance = (int)glyph.Advance.X.Value,
					Texture = t,
					AdvanceY = glyph.Advance.Y.Value
				};

				Label._characters[con][l].Add(temp, cha);
				return t;
				
			}

			if (!BadChars.ContainsKey(l))
			{
				BadChars.Add(l, TextureManager.TextureManagers[con].AddTexture(Tools.GetResourceStream("GraphicsManager.Resources.Textures.BadChar.png")));
				BadChars[l].Unit = (s.GetUniformLocation("u_texture") switch
				{
					0 => TextureUnit.Texture0,
					1 => TextureUnit.Texture1,
					2 => TextureUnit.Texture2,
					3 => TextureUnit.Texture3,
					4 => TextureUnit.Texture4,
					5 => TextureUnit.Texture5,
					6 => TextureUnit.Texture6,
					7 => TextureUnit.Texture7,
					8 => TextureUnit.Texture8,
					9 => TextureUnit.Texture9,
				});
				Label._characters[con][l].Add(temp, new()
				{
					Size = Label._characters[con][l].First().Value.Size,
					Bearing = Label._characters[con][l].First().Value.Bearing,
					Advance = Label._characters[con][l].First().Value.Advance,
					Texture = BadChars[l],
					AdvanceY = Label._characters[con][l].First().Value.AdvanceY
				});
				if (TextureManager.TextureManagers[con].Window.ShowMissingChar)
					Console.WriteLine($"Missing font for '{charter}:{(uint)charter}' in Family {l.Families[0].Family}");
			}
			
			return BadChars[l];
		}
		catch (Exception e)
		{
			Console.WriteLine(e);
			throw;
		}
	}

	public Texture Clone()
	{
		return (Texture)MemberwiseClone();
	}

	public TextureUnit Unit = TextureUnit.Texture0;

	public void Use()
	{
		GL.ActiveTexture(Unit);
		GL.BindTexture(TextureTarget.Texture2D, handel);
	}
}
