﻿using System.Diagnostics;
using GraphicsManager.Enums;
using GraphicsManager.Interfaces;
using GraphicsManager.Objects.Core;
using OpenTK.Graphics.OpenGL4;
using OpenTK.Mathematics;
using OpenTK.Windowing.Common;
using OpenTK.Windowing.Common.Input;
using OpenTK.Windowing.Desktop;
using OpenTK.Windowing.GraphicsLibraryFramework;

namespace GraphicsManager.Objects;

public class Rectangle : ITextureObject
{
	public static readonly Dictionary<IGLFWGraphicsContext, Shader> DefaultShader = new();
	public static readonly Dictionary<IGLFWGraphicsContext, Shader> DefaultAlphaShader = new();
	public static readonly Dictionary<IGLFWGraphicsContext, Shader> DefaultAlphaTextureShader = new();

	public ObjectAnchor Anchor { get; set; } = ObjectAnchor.Left | ObjectAnchor.Top;

	public Vector3i GetWindowLocation()
	{
		if (!Loaded) return Location;
		Vector3i loc = Location;
		IParent? p = Parent;
		while (p is not null)
		{
			loc += p.Position;
			p = p.Parent;
		}

		return loc;
	}
	
	public Vector3i GetParentLocation(IParent par)
	{
		if (!Loaded) return Location;
		Vector3i loc = Location;
		IParent? p = Parent;
		while (p is not null && p != par)
		{
			loc += p.Position;
			p = p.Parent;
		}

		return loc;
	}

	public List<Texture> Textures { get; set; } = new();
	public BetterContextMenu? ContextMenu { get; set; } = null;
	public event Func<string[], Task>? FilesDroped; 

	public Rectangle(Texture? texture)
	{
		if (texture is not null) Textures.Add(texture);;
	}
	public Rectangle()
	{
	}
	
	public virtual void Focus()
	{
		
	}
	public virtual void UnFocus()
	{
		
	}

	private Color4 _BackgroundColor { get; set; } = new(0, 0, 0, 255);

	public virtual Color4 BackgroundColor
	{
		get => _BackgroundColor;
		set
		{
			_BackgroundColor = value;
			if (Parent is not null) Parent.TryDraw();
		}
	}

	private bool _Visible = true;
	public virtual bool Visible
	{
		get => _Visible;
		set
		{
			_Visible = value; 
			if (Parent is not null) Parent.TryDraw();
		}
	}

	public Action? OnDrawAction;
	//public Action? OnDrawExitAction;

	public virtual void Draw(int x, int y, int w, int h)
	{
		if (Visible && Loaded && Location.X > 0 - Size.X && Location.Y > 0 - Size.Y)
		{
			if (OnDrawAction is not null) OnDrawAction.Invoke();
			if (!Window!.Context.IsCurrent) Window.Context.MakeCurrent();
			foreach (Texture tex in Textures)
			{
				tex.Use();
			}
			Shader.Use();
			if (!Textures.Any() || Shader.Handle == DefaultAlphaShader[Window!.Context].Handle)
			{ 
				GL.Uniform4(0, BackgroundColor);
			}
			GL.BindVertexArray(ArrayObject);
			GL.DrawElements(PrimitiveType.Triangles, Indexs.Length, DrawElementsType.UnsignedShort, 0);
		}
	}
	
	public virtual MouseCursor HoverMouse { get; set; } = MouseCursor.Default;

	public void ForceDistanceUpdate()
	{
		if (Parent is null) return;
		ForceDistanceUpdate(Parent);
	}
	
	public void ForceDistanceUpdate(IParent parent)
	{
		Distance = new(parent.Size.X - Size.X - Location.X, parent.Size.Y - Size.Y - Location.Y);
	}

	public virtual void Clean()
	{
		GL.BindBuffer(BufferTarget.ArrayBuffer, 0);
		GL.DeleteBuffer(BufferObject);
		GL.DeleteBuffer(ElementBufferObject);
		GL.DeleteVertexArray(ArrayObject);
	}

	public virtual void LoadToParent(IParent Parent, IWindow Window)
	{
		if (Loaded) return;
		this.Parent = Parent;
		this.Window = Window;
		if (Shader is null)
		{
			if (!Textures.Any())Shader = DefaultShader[Window.Context];
			else Shader = Texture.TextureShader[Window.Context];
		}
		BufferObject = GL.GenBuffer();
		ArrayObject = GL.GenVertexArray();
		if (Textures.Any() && !Shader.ForTexture)
		{
			Shader = Texture.TextureShader[Window.Context];
		}
		ElementBufferObject = GL.GenBuffer();
		GL.BindBuffer(BufferTarget.ElementArrayBuffer, ElementBufferObject);
		GL.BufferData(BufferTarget.ElementArrayBuffer, Indexs.Length * sizeof(ushort), Indexs, Hint);
		Loaded = true;
		Window.MouseDown += Window_MouseDown;
		Window.FileDrop += WindowOnFileDrop;
		Location = Location;
		if (Distance.X == 0 && Distance.Y == 0) Distance = new(Parent.Size.X - Size.X - Location.X, Parent.Size.Y - Size.Y - Location.Y);
		if (WindowLoaded is not null) WindowLoaded.Invoke(this);
	}

	public void TransferOwners(IWindow w, IParent p)
	{
		if (Parent is not null)
		{
			Parent.Controls._internal.Remove(this);
			Window = w;
			Parent = p;
		}
		p.Controls.Add(this);
	}

	private void WindowOnFileDrop(FileDropEventArgs obj)
	{
		if (!MouseInside) return;
		if (FilesDroped is not null) _ = FilesDroped.Invoke(obj.FileNames);
	}

	private bool mi = false;

	public bool MouseInside
	{
		get => mi;
		set
		{
			if (Window is null) return;
			if (AllowHoverFromBehind)
			{
				mi = value;
				if (value && MouseEnter is not null)  MouseEnter.Invoke(this);
				if (!value && MouseLeave is not null)  MouseLeave.Invoke(this);
			}
			else
			{
				if (Window.HoveringControl == this && value)
				{
					mi = value;
					if (MouseEnter is not null) MouseEnter.Invoke(this);
				}

				if (Window.HoveringControl != this && !value)
				{
					mi = value;
					if (MouseLeave is not null) MouseLeave.Invoke(this);
				}
			}
		}
	}

	public IParent? Parent { get; private set; }
	public IWindow? Window { get; private set; }
	private bool ahfb = false;

	public bool AllowHoverFromBehind
	{
		get
		{
			return ahfb;
		}
		set
		{
			ahfb = value;
			if (!value)
			{
				if (MouseInside && Window is not null && Window.HoveringControl != this)
				{
					mi = false;
					if (MouseLeave is not null) MouseLeave.Invoke(this);
				}
			}
		}
	}

	private void Window_MouseDown(OpenTK.Windowing.Common.MouseButtonEventArgs e)
	{
		if (MouseInside && e.Button == MouseButton.Button1 && Clicked is not null) _ = Clicked.Invoke(this);
		if (MouseInside && e.Button == MouseButton.Button2 && ContextMenu is not null && Window!.HoveringControl == this) ContextMenu.ShowContext(Window!);
		if (e.Button != MouseButton.Button2 && ContextMenu is not null && ContextMenu.Visible) ContextMenu.HideContext(Window!);
	}

	~Rectangle()
	{
		GL.BindBuffer(BufferTarget.ArrayBuffer, 0);
		GL.DeleteBuffer(BufferObject);
		GL.DeleteBuffer(ElementBufferObject);
		GL.DeleteVertexArray(ArrayObject);
	}

	public Shader Shader { get; set; } = null;
	public int ElementBufferObject { get; private set; }
	public int BufferObject { get; private set; }
	public bool IgnoreHover { get; set; }
	public int ArrayObject { get; private set; }
	private Vector2i size_ = new();
	private Vector3i loc_ = new();

	public float[] Points
	{
		set
		{
			try
			{
				if (Loaded)
				{
					
					GL.BindBuffer(BufferTarget.ArrayBuffer, BufferObject);
					GL.BindVertexArray(ArrayObject);
					int add = 3;
					if (Textures.Any())
					{
						add = 5;
						GL.EnableVertexAttribArray(Textures.First().Location);
						GL.VertexAttribPointer(Textures.First().Location, 2, VertexAttribPointerType.Float, false, 5 * sizeof(float), 3 * sizeof(float));
						
					}
					GL.VertexAttribPointer(0, 3, VertexAttribPointerType.Float, false, add * sizeof(float), 0);
					GL.EnableVertexAttribArray(0);
					GL.BindBuffer(BufferTarget.ArrayBuffer, BufferObject);
					GL.BufferData(BufferTarget.ArrayBuffer, value.Length * sizeof(float), value, Hint);
					GL.BindVertexArray(ArrayObject);
					GL.BindBuffer(BufferTarget.ElementArrayBuffer, ElementBufferObject);
					GL.BufferData(BufferTarget.ElementArrayBuffer, Indexs.Length * sizeof(ushort), Indexs, Hint);
					if (Window is not null && Window.CanControleUpdate && Loaded) Parent!.TryDraw();
				}
			}
			catch (AccessViolationException v)
			{
				Console.WriteLine(v.Message);
			}
		}
	}
	
	public event Func<IRenderObject, Task>? SizeChanged; 

	public ushort[] Indexs { get; set; } = new ushort[6] { 0, 1, 3, 1, 2, 3 };
	public BufferUsageHint Hint { get; set; } = BufferUsageHint.StaticDraw;
	public event Func<IRenderObject, Task>? Clicked;
	public bool Loaded { get; private set; }
	public event Func<IRenderObject, Task>? WindowLoaded;
	public event Func<IRenderObject, Task>? MouseEnter;
	public event Func<IRenderObject, Task>? MouseLeave;
	public object? Tag { get; set; }
	public Vector2i Distance { get; set; }

	private TextureDisplay td = TextureDisplay.Clamped;

	public TextureDisplay TextureDisplay
	{
		get
		{
			return td;
		}
		set
		{
			if (Textures.Count == 0) return;
			TextureDisplay old = td;
			td = value;
			float per = (float)Textures[0].RawSize!.Value.Y / Textures[0].RawSize!.Value.X;
			float diff = 0;
			switch (old)
			{
				default:
					switch (value)
					{
						case TextureDisplay.Clamped:
							Points = new float[]
							{
								saf.X,        laf.Y, 0, Textures[0].MaxText.X,     Textures[0].MaxText.Y,
								saf.X,        saf.Y, 0, Textures[0].MaxText.X,     0,
								laf.X,        saf.Y, 0, 0,                         0,
								laf.X,        laf.Y, 0, 0,                         Textures[0].MaxText.Y,
							};
							break;
						case TextureDisplay.HorizontalCenter or TextureDisplay.ProgressHorizontalCenter or TextureDisplay.TextureHorizontalCenter:
							
							if (Window is not null)
							{
								diff = Window.IntToFloat(Size.Y) + 1;
								if (value == TextureDisplay.TextureHorizontalCenter)
								{
									diff = Window.IntToFloat(Textures[0].RawSize!.Value.Y);
								}
							}
							Points = new float[]
							{
								saf.X,        laf.Y, 0, Textures[0].MaxText.X,     Textures[0].MaxText.Y,	//4
								saf.X,        saf.Y, 0, Textures[0].MaxText.X,     0,						//9
								laf.X,        saf.Y, 0, 0,                         0,						//14
								laf.X,        laf.Y, 0, 0,                         Textures[0].MaxText.Y,	//19
								saf.X - diff, laf.Y, 0, Textures[0].MaxText.X-per, Textures[0].MaxText.Y,	//24
								saf.X - diff, saf.Y, 0, Textures[0].MaxText.X-per, 0,						//29
								laf.X + diff, saf.Y, 0, per,                       0,						//34
								laf.X + diff, laf.Y, 0, per,                       Textures[0].MaxText.Y,	//39
							};
							Indexs = new ushort[]
							{
								0, 1, 4,
								1, 4, 5,
								4, 5, 6,
								4, 6, 7,
								3, 6, 7,
								2, 3, 6
							};
							break;
						
						case TextureDisplay.Center:
							per = (float)Textures[0].RawSize!.Value.X / 3;;
							diff = 0;
							if (Window is not null)
							{
								if (Size.X > Textures[0].RawSize!.Value.X)
									diff = (Window.IntToFloat(Textures[0].RawSize!.Value.X) + 1) / 3;
								else
									diff = (Window.IntToFloat(Size.X) + 1) / 3;
							}
							Points = new float[]
							{
								saf.X,        laf.Y, 0, Textures[0].MaxText.X,     Textures[0].MaxText.Y,	//4
								saf.X,        saf.Y, 0, Textures[0].MaxText.X,     0,						//9
								laf.X,        saf.Y, 0, 0,                         0,						//14
								laf.X,        laf.Y, 0, 0,                         Textures[0].MaxText.Y,	//19
								saf.X - diff, laf.Y, 0, Textures[0].MaxText.X-per, Textures[0].MaxText.Y,	//24
								saf.X - diff, saf.Y, 0, Textures[0].MaxText.X-per, 0,						//29
								laf.X + diff, saf.Y, 0, per,                       0,						//34
								laf.X + diff, laf.Y, 0, per,                       Textures[0].MaxText.Y,	//39
								saf.X - diff, laf.Y, 0, Textures[0].MaxText.X-per, Textures[0].MaxText.Y,	//24
								saf.X - diff, saf.Y, 0, Textures[0].MaxText.X-per, 0,						//29
								laf.X + diff, saf.Y, 0, per,                       0,						//34
								laf.X + diff, laf.Y, 0, per,                       Textures[0].MaxText.Y,	//39
								saf.X - diff, laf.Y, 0, Textures[0].MaxText.X-per, Textures[0].MaxText.Y,	//24
								saf.X - diff, saf.Y, 0, Textures[0].MaxText.X-per, 0,						//29
								laf.X + diff, saf.Y, 0, per,                       0,						//34
								laf.X + diff, laf.Y, 0, per,                       Textures[0].MaxText.Y,	//39
							};
							Indexs = new ushort[]
							{
								0,  4,  11,
								4,  11, 12,
								4,  5,  12,
								5,  12, 13,
								1,  5,  6,
								5,  6,  13,
								10, 11, 12,
								10, 12, 15,
								12, 13, 14,
								12, 14, 15,
								6,  7,  13,
								7,  13, 14,
								3,  9,  10,
								9,  10, 15,
								8,  9,  14,
								9,  14, 15,
								2,  7,  8,
								7,  8,  14,
							};
							break;
					}
					break;
			}
			if (Parent is not null) Parent.TryDraw();
		}
	}
	
	private static void SetPoint(float[] temp, int start, float X, float Y, float Z, float Tx, float Ty)
	{
		temp[start] = X;
		temp[start + 1] = Y;
		temp[start + 2] = Z;
		temp[start + 3] = Tx;
		temp[start + 4] = Ty;
	}
	
	public virtual Vector2i Size
	{
		get
		{
			return size_;
		}
		set
		{
			size_ = value;
			if (Window is null || Parent is null) return;
			Parent.ReportSizeUpdate(this);
			//float[] temp = Points;
			float[] temp = Textures.Count == 0
				? new float[12]
				: TextureDisplay switch
				{
					TextureDisplay.Clamped => new float[20],
					TextureDisplay.HorizontalCenter or TextureDisplay.ProgressHorizontalCenter or TextureDisplay.TextureHorizontalCenter => new float[40],
					TextureDisplay.Center => new float[80],
					_ => new float[20]
				};
			saf = new Vector2(Parent.IntToFloat(value.X + loc_.X, false), Parent.IntToFloat(value.Y + loc_.Y, true));
			temp[2] = Location.Z;
			temp[(!Textures.Any() ? 5 : 7)] = Location.Z;
			temp[(!Textures.Any() ? 8 : 12)] = Location.Z;
			temp[(!Textures.Any() ? 11 : 17)] = Location.Z;
			
			temp[0] = saf.X; // top r
			temp[1] = laf.Y;
			temp[(!Textures.Any()? 3 : 5)] = saf.X; // bot r
			temp[(!Textures.Any() ? 4 : 6)] = saf.Y;
			temp[(!Textures.Any() ? 6 : 10)] = laf.X;//bot l
			temp[(!Textures.Any() ? 7 : 11)] = saf.Y;
			temp[(!Textures.Any() ? 9 : 15)] = laf.X;//top l
			temp[(!Textures.Any() ? 10 : 16)] = laf.Y;

			if (Textures.Count > 0)
			{
				Vector2i s = value;
				float diff = Window.IntToFloat(s.Y) + 1;
				float per = (float)Textures[0].RawSize!.Value.Y / Textures[0].RawSize!.Value.X;
				if (TextureDisplay == TextureDisplay.TextureHorizontalCenter)
					diff = Window.IntToFloat(Textures[0].RawSize!.Value.Y);
				if (TextureDisplay == TextureDisplay.Center)
				{
					if (s.X > Textures[0].RawSize!.Value.X)
						diff = (Window.IntToFloat(Textures[0].RawSize!.Value.X) + 1) / 3;
					else
						diff = (Window.IntToFloat(s.X) + 1) / 3;
					per = Textures[0].MaxText.X / 3;
				}

				switch (TextureDisplay)
				{
					case TextureDisplay.Clamped:
						temp[3] = Textures[0].MaxText.X;
						temp[4] = Textures[0].MaxText.Y;
						temp[8] = Textures[0].MaxText.X;
						temp[19] = Textures[0].MaxText.Y;
						break;
					case TextureDisplay.HorizontalCenter:
						temp[3] = Textures[0].MaxText.X;
						temp[4] = Textures[0].MaxText.Y;
						temp[8] = Textures[0].MaxText.X;
						temp[19] = Textures[0].MaxText.Y;
						temp[20] = saf.X - diff; // top r
						temp[21] = laf.Y;
						temp[22] = 0;
						temp[23] = Textures[0].MaxText.X - per;
						temp[24] = 1;
						
						temp[25] = saf.X - diff; // bot r
						temp[26] = saf.Y;
						temp[27] = 0;
						temp[28] = Textures[0].MaxText.X - per;
						temp[29] = 0;
						
						temp[30] = laf.X + diff; // bot l
						temp[31] = saf.Y;
						temp[32] = 0;
						temp[33] = per;
						temp[34] = 0;
						
						temp[35] = laf.X + diff; // top l
						temp[36] = laf.Y;
						temp[37] = 0;
						temp[38] = per;
						temp[39] = 1;
						break;
					case TextureDisplay.Center:
						temp[3] = Textures[0].MaxText.X;
						temp[4] = Textures[0].MaxText.Y;
						temp[8] = Textures[0].MaxText.X;
						temp[19] = Textures[0].MaxText.Y;

						int dy, dx;
						if (s.Y > Textures[0].RawSize!.Value.Y)
							dy = Textures[0].RawSize!.Value.Y;
						else
							dy = s.Y;
						if (s.X > Textures[0].RawSize!.Value.X)
							dx = Textures[0].RawSize!.Value.X;
						else
							dx = s.X;

						if (dy > dx) dy = dx;
						else dx = dy;
						
						float diffy =  (Window.IntToFloat(Window.Size.Y - dy, true) + 1) / 3;
						diff =  (Window.IntToFloat(dx) + 1) / 3;
						
						per = Textures[0].MaxText.X / 3;
						
						SetPoint(temp, 20, temp[0],         temp[1] - diffy, temp[2], temp[3],                 temp[4] - (temp[4]*per));
						SetPoint(temp, 25, temp[0],         temp[6] + diffy, temp[2], temp[3],                 temp[4] * per);
						SetPoint(temp, 30, temp[5] - diff,  temp[6],         temp[2], temp[3] - (temp[3]*per), temp[9]);
						SetPoint(temp, 35, temp[10] + diff, temp[6],         temp[2], temp[3] * per,           temp[9]);
						SetPoint(temp, 40, temp[10],        temp[26],        temp[2], temp[13],                temp[29]);
						SetPoint(temp, 45, temp[10],        temp[21],        temp[2], temp[13],                temp[24]);
						SetPoint(temp, 50, temp[35],        temp[1],         temp[2], temp[38],                temp[4]);
						SetPoint(temp, 55, temp[30],        temp[1],         temp[2], temp[33],                temp[4]);
						SetPoint(temp, 60, temp[30],        temp[21],        temp[2], temp[33],                temp[24]);
						SetPoint(temp, 65, temp[30],        temp[26],        temp[2], temp[33],                temp[29]);
						SetPoint(temp, 70, temp[35],        temp[26],        temp[2], temp[38],                temp[29]);
						SetPoint(temp, 75, temp[35],        temp[21],        temp[2], temp[38],                temp[24]);
						break;
					case TextureDisplay.ProgressHorizontalCenter:
						if (s.X > s.Y)
						{
							if (s.X > s.Y + s.Y)
							{
								temp[3] = Textures[0].MaxText.X;
								temp[4] = Textures[0].MaxText.Y;
								temp[8] = Textures[0].MaxText.X;
								temp[19] = Textures[0].MaxText.Y;
								temp[20] = saf.X - diff; // top r
								temp[21] = laf.Y;
								temp[22] = 0;
								temp[23] = Textures[0].MaxText.X - per;
								temp[24] = 1;
						
								temp[25] = saf.X - diff; // bot r
								temp[26] = saf.Y;
								temp[27] = 0;
								temp[28] = 1 - per;
								temp[29] = 0;
						
								temp[30] = laf.X + diff; // bot l
								temp[31] = saf.Y;
								temp[32] = 0;
								temp[33] = per;
								temp[34] = 0;
						
								temp[35] = laf.X + diff; // top l
								temp[36] = laf.Y;
								temp[37] = 0;
								temp[38] = per;
								temp[39] = 1;
							}
							else
							{
								temp[3] = (0.666666f + (((s.X - s.Y) / (float)s.Y)* per)) * Textures[0].MaxText.X;
								temp[4] = Textures[0].MaxText.Y;
								temp[8] = temp[3];
								temp[19] = Textures[0].MaxText.Y;
								//start last 33% of texture
								temp[20] = saf.X; // top r
								temp[21] = laf.Y;
								temp[22] = 0;
								temp[23] = temp[3];
								temp[24] = 1;

								temp[25] = saf.X; // bot r
								temp[26] = saf.Y;
								temp[27] = 0;
								temp[28] = temp[3];
								temp[29] = 0;
						
								temp[30] = laf.X + diff; // bot l
								temp[31] = saf.Y;
								temp[32] = 0;
								temp[33] = per;
								temp[34] = 0;
						
								temp[35] = laf.X + diff; // top l
								temp[36] = laf.Y;
								temp[37] = 0;
								temp[38] = per;
								temp[39] = 1;
							}
						}
						else
						{
							//first 33% of texture
							temp[3] = ((s.X/(float)s.Y)*per);
							temp[4] = Textures[0].MaxText.Y;
							temp[8] = temp[3];
							temp[19] = Textures[0].MaxText.Y;
							temp[20] = saf.X; // top r
							temp[21] = laf.Y;
							temp[22] = 0;
							temp[23] = temp[3];
							temp[24] = Textures[0].MaxText.Y;
						
							temp[25] = saf.X; // bot r
							temp[26] = saf.Y;
							temp[27] = 0;
							temp[28] = temp[3];
							temp[29] = 0;
						
							temp[30] = saf.X; // bot l
							temp[31] = saf.Y;
							temp[32] = 0;
							temp[33] = temp[3];
							temp[34] = 0;
						
							temp[35] = saf.X; // top l
							temp[36] = laf.Y;
							temp[37] = 0;
							temp[38] = temp[3];
							temp[39] = Textures[0].MaxText.Y;
						}
						break;
				}
			}
			
			Points = temp;
			if (SizeChanged is not null) SizeChanged.Invoke(this);
		}
	}
	
	public virtual Vector3i Location
	{
		get
		{
			return loc_;
		}
		set
		{
			loc_ = value;
			if (Window is null || Parent is null) return;
			float[] temp = Textures.Count == 0
				? new float[12]
				: TextureDisplay switch
				{
					TextureDisplay.Clamped => new float[20],
					TextureDisplay.HorizontalCenter or TextureDisplay.ProgressHorizontalCenter => new float[40],
					TextureDisplay.Center => new float[80],
					_ => new float[20]
				};
			laf = new Vector2(Parent.IntToFloat(value.X, false), Parent.IntToFloat(value.Y, true));
			temp[2] = value.Z;
			temp[(!Textures.Any() ? 5 : 7)] = value.Z;
			temp[(!Textures.Any() ? 8 : 12)] = value.Z;
			temp[(!Textures.Any() ? 11 : 17)] = value.Z;
			
			temp[(!Textures.Any() ? 6 : 10)] = laf.X;
			temp[(!Textures.Any() ? 9 : 15)] = laf.X;
			temp[1] = laf.Y;
			temp[(!Textures.Any() ? 10 : 16)] = laf.Y;
			saf = new Vector2(Parent.IntToFloat(Size.X + value.X, false), Parent.IntToFloat(Size.Y + value.Y, true));
			temp[0] = saf.X;
			temp[(!Textures.Any() ? 3 : 5)] = saf.X;
			temp[(!Textures.Any() ? 4 : 6)] = saf.Y;
			temp[(!Textures.Any() ? 7 : 11)] = saf.Y;
			if (Textures.Count > 0)
			{
				Vector2i s = Size;
				float diff = Window.IntToFloat(s.Y) + 1;
				float per = (float)Textures[0].RawSize!.Value.Y / Textures[0].RawSize!.Value.X;
				if (TextureDisplay == TextureDisplay.Center)
				{
					if (s.X > Textures[0].RawSize!.Value.X)
						diff = (Window.IntToFloat(Textures[0].RawSize!.Value.X) + 1) / 3;
					else
						diff = (Window.IntToFloat(s.X) + 1) / 3;
					per = Textures[0].MaxText.X / 3;
				}
				switch (TextureDisplay)
				{
					case TextureDisplay.Clamped:
						temp[3] = Textures[0].MaxText.X;
						temp[4] = Textures[0].MaxText.Y;
						temp[8] = Textures[0].MaxText.X;
						temp[19] = Textures[0].MaxText.Y;
						break;
					case TextureDisplay.HorizontalCenter:
						temp[3] = Textures[0].MaxText.X;
						temp[4] = Textures[0].MaxText.Y;
						temp[8] = Textures[0].MaxText.X;
						temp[19] = Textures[0].MaxText.Y;
						temp[20] = saf.X - diff; // top r
						temp[21] = laf.Y;
						temp[22] = 0;
						temp[23] = Textures[0].MaxText.X - per;
						temp[24] = 1;
						
						temp[25] = saf.X - diff; // bot r
						temp[26] = saf.Y;
						temp[27] = 0;
						temp[28] = Textures[0].MaxText.X - per;
						temp[29] = 0;
						
						temp[30] = laf.X + diff; // bot l
						temp[31] = saf.Y;
						temp[32] = 0;
						temp[33] = per;
						temp[34] = 0;
						
						temp[35] = laf.X + diff; // top l
						temp[36] = laf.Y;
						temp[37] = 0;
						temp[38] = per;
						temp[39] = 1;
						break;
					case TextureDisplay.Center:
						temp[3] = Textures[0].MaxText.X;
						temp[4] = Textures[0].MaxText.Y;
						temp[8] = Textures[0].MaxText.X;
						temp[19] = Textures[0].MaxText.Y;
						
						int dy, dx;
						if (s.Y > Textures[0].RawSize!.Value.Y)
							dy = Textures[0].RawSize!.Value.Y;
						else
							dy = s.Y;
						if (s.X > Textures[0].RawSize!.Value.X)
							dx = Textures[0].RawSize!.Value.X;
						else
							dx = s.X;

						if (dy > dx) dy = dx;
						else dx = dy;
						
						float diffy =  (Window.IntToFloat(Window.Size.Y - dy, true) + 1) / 3;
						diff =  (Window.IntToFloat(dx) + 1) / 3;
						
						SetPoint(temp, 20, temp[0],         temp[1] - diffy, temp[2], temp[3],                 temp[4] - (temp[4]*per));
						SetPoint(temp, 25, temp[0],         temp[6] + diffy, temp[2], temp[3],                 temp[4] * per);
						SetPoint(temp, 30, temp[5] - diff,  temp[6],         temp[2], temp[3] - (temp[3]*per), temp[9]);
						SetPoint(temp, 35, temp[10] + diff, temp[6],         temp[2], temp[3] * per,           temp[9]);
						SetPoint(temp, 40, temp[10],        temp[26],        temp[2], temp[13],                temp[29]);
						SetPoint(temp, 45, temp[10],        temp[21],        temp[2], temp[13],                temp[24]);
						SetPoint(temp, 50, temp[35],        temp[1],         temp[2], temp[38],                temp[4]);
						SetPoint(temp, 55, temp[30],        temp[1],         temp[2], temp[33],                temp[4]);
						SetPoint(temp, 60, temp[30],        temp[21],        temp[2], temp[33],                temp[24]);
						SetPoint(temp, 65, temp[30],        temp[26],        temp[2], temp[33],                temp[29]);
						SetPoint(temp, 70, temp[35],        temp[26],        temp[2], temp[38],                temp[29]);
						SetPoint(temp, 75, temp[35],        temp[21],        temp[2], temp[38],                temp[24]);
						break;
					case TextureDisplay.ProgressHorizontalCenter:
						if (s.X > s.Y)
						{
							if (s.X > s.Y + s.Y)
							{
								temp[3] = Textures[0].MaxText.X;
								temp[4] = Textures[0].MaxText.Y;
								temp[8] = Textures[0].MaxText.X;
								temp[19] = Textures[0].MaxText.Y;
								temp[20] = saf.X - diff; // top r
								temp[21] = laf.Y;
								temp[22] = 0;
								temp[23] = Textures[0].MaxText.X - per;
								temp[24] = 1;
						
								temp[25] = saf.X - diff; // bot r
								temp[26] = saf.Y;
								temp[27] = 0;
								temp[28] = 1 - per;
								temp[29] = 0;
						
								temp[30] = laf.X + diff; // bot l
								temp[31] = saf.Y;
								temp[32] = 0;
								temp[33] = per;
								temp[34] = 0;
						
								temp[35] = laf.X + diff; // top l
								temp[36] = laf.Y;
								temp[37] = 0;
								temp[38] = per;
								temp[39] = 1;
							}
							else
							{
								temp[3] = (0.666666f + (((s.X - s.Y) / (float)s.Y)* per)) * Textures[0].MaxText.X;
								temp[4] = Textures[0].MaxText.Y;
								temp[8] = temp[3];
								temp[19] = Textures[0].MaxText.Y;
								//start last 33% of texture
								temp[20] = saf.X; // top r
								temp[21] = laf.Y;
								temp[22] = 0;
								temp[23] = temp[3];
								temp[24] = 1;

								temp[25] = saf.X; // bot r
								temp[26] = saf.Y;
								temp[27] = 0;
								temp[28] = temp[3];
								temp[29] = 0;
						
								temp[30] = laf.X + diff; // bot l
								temp[31] = saf.Y;
								temp[32] = 0;
								temp[33] = per;
								temp[34] = 0;
						
								temp[35] = laf.X + diff; // top l
								temp[36] = laf.Y;
								temp[37] = 0;
								temp[38] = per;
								temp[39] = 1;
							}
						}
						else
						{
							//first 33% of texture
							temp[3] = ((s.X/(float)s.Y)*per);
							temp[4] = Textures[0].MaxText.Y;
							temp[8] = temp[3];
							temp[19] = Textures[0].MaxText.Y;
							temp[20] = saf.X; // top r
							temp[21] = laf.Y;
							temp[22] = 0;
							temp[23] = temp[3];
							temp[24] = Textures[0].MaxText.Y;
						
							temp[25] = saf.X; // bot r
							temp[26] = saf.Y;
							temp[27] = 0;
							temp[28] = temp[3];
							temp[29] = 0;
						
							temp[30] = saf.X; // bot l
							temp[31] = saf.Y;
							temp[32] = 0;
							temp[33] = temp[3];
							temp[34] = 0;
						
							temp[35] = saf.X; // top l
							temp[36] = laf.Y;
							temp[37] = 0;
							temp[38] = temp[3];
							temp[39] = Textures[0].MaxText.Y;
						}
						break;
				}
			}
			Points = temp;
		}
	}
	
	private Vector2 laf = new(), saf = new();

	public Vector2 LocationAsFloat { get { return laf; } }
	public Vector2 SizeAsFloat { get { return saf; } }
}
